@@ -2,7 +2,7 @@
 
 
   Authors: Anton Ertl, Bernd Paysan, Jens Wilke, David Kühling
-  Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019,2020,2021 Free Software Foundation, Inc.
+  Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019,2020,2021,2022 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -198,6 +198,7 @@ static int print_metrics=0; /* if true, print metrics on exit */
 static int print_prims=0; /* if true, print primitives on exit */
 static int static_super_number = 10000; /* number of ss used if available */
 #define MAX_STATE 9 /* maximum number of states */
+#define CANONICAL_STATE 0
 static int maxstates = MAX_STATE; /* number of states for stack caching */
 static int ss_greedy = 0; /* if true: use greedy, not optimal ss selection */
 static int tpa_noequiv = 0;     /* if true: no state equivalence checking */
@@ -249,10 +250,6 @@ typedef struct {
 PrimInfo *priminfos;
 PrimInfo **decomp_prims;
 
-const char * const prim_names[]={
-#include PRIM_NAMES_I
-};
-
 void init_ss_cost(void);
 
 static int is_relocatable(int p)
@@ -266,6 +263,12 @@ static int is_relocatable(int p)
 }
 #endif /* defined(NO_DYNAMIC) */
 
+const char * const prim_names[]={
+#if !defined(NO_DYNAMIC)
+#include PRIM_NAMES_I
+#endif
+};
+
 #ifdef MEMCMP_AS_SUBROUTINE
 int gforth_memcmp(const char * s1, const char * s2, size_t n)
 {
@@ -966,14 +969,14 @@ MAYBE_UNUSED static Label bsearch_next(Label key, Label *a, UCell n)
     return bsearch_next(key, a, mid+1);
 }
 
-#ifndef NO_DYNAMIC
-static int state_map(int state)
+int state_map(int state)
 {
   if (state==0) return STACK_CACHE_DEFAULT;
   if (state<=STACK_CACHE_DEFAULT) return state-1;
   return state;
 }
 
+#ifndef NO_DYNAMIC
 static void gforth_printprims()
 {
   unsigned i;
@@ -1135,6 +1138,54 @@ static void check_prims(Label symbols1[])
 #endif
 }
 
+/* Dynamic info for decompilation */
+DynamicInfo *dynamicinfos = NULL; /* 2^n-sized growable array */
+long ndynamicinfos=0; /* index of next dynamicinfos entry */
+
+DynamicInfo *dynamic_info(Label code)
+{
+  DynamicInfo *di;
+#if 0
+  /* !! faster implementation: linear search for the block, then use
+        binary search within the block */
+  struct code_block_list *p;
+  Address code=_code;
+
+  /* first, check if we are in code at all */
+  for (p = code_block_list;; p = p->next) {
+    if (p == NULL)
+      return code;
+    if (code >= p->block && code < p->block+p->size)
+      break;
+  }
+#endif
+  for (di=dynamicinfos; di<&dynamicinfos[ndynamicinfos]; di++)
+    if (di->start == code)
+      return di;
+  return NULL;
+}
+
+#if !(defined(DOUBLY_INDIRECT) || defined(INDIRECT_THREADED))
+static DynamicInfo *add_dynamic_info()
+/* reserves space for a new Dynamicinfo, returning a pointer to it (for
+   filling out) */
+{
+  long old=ndynamicinfos;
+  long new=old+1;
+  if (dynamicinfos!=NULL && dynamicinfos[old-1].length == 0) {
+    return &dynamicinfos[old-1];
+  }
+  ndynamicinfos=new;
+  if ((old&new)==0) { /* one too early for old>=1, but we can live with that */
+    dynamicinfos =
+      (DynamicInfo *)realloc(dynamicinfos,2*new*sizeof(DynamicInfo));
+    if (dynamicinfos==NULL)
+      perror(progname);
+  }
+  return &dynamicinfos[old];
+}
+#endif
+
 static void flush_to_here(void)
 {
 #ifndef NO_DYNAMIC
@@ -1144,20 +1195,24 @@ static void flush_to_here(void)
 #endif
 }
 
+static void MAYBE_UNUSED  append_code(Address code, size_t length)
+{
+  memmove(code_here,code,length);
+  code_here += length;
+}
+
 static void MAYBE_UNUSED align_code(void)
      /* align code_here on some platforms */
 {
 #ifndef NO_DYNAMIC
 #if defined(CODE_PADDING)
   Cell alignment = CODE_ALIGNMENT;
-  static char nops[] = CODE_PADDING;
+  static unsigned char nops[] = CODE_PADDING;
   UCell maxpadding=MAX_PADDING;
   UCell offset = ((UCell)code_here)&(alignment-1);
   UCell length = alignment-offset;
-  if (length <= maxpadding) {
-    memmove(code_here,nops+offset,length);
-    code_here += length;
-  }
+  if (length <= maxpadding)
+    append_code(nops+offset,length);
 #endif /* defined(CODE_PADDING) */
 #endif /* defined(NO_DYNAMIC */
 }  
@@ -1167,18 +1222,21 @@ static void append_jump(void)
 {
   if (last_jump) {
     PrimInfo *pi = &priminfos[last_jump];
-    
     /* debugp(stderr, "Copy code %p<=%p+%x,%d\n", code_here, pi->start, pi->length, pi->restlength); */
-    memmove(code_here, pi->start+pi->length, pi->restlength);
-    code_here += pi->restlength;
+    append_code(pi->start+pi->length, pi->restlength);
     /* debugp(stderr, "Copy goto %p<=%p,%d\n", code_here, goto_start, goto_len); */
-    memmove(code_here, goto_start, goto_len);
-    code_here += goto_len;
+    append_code(goto_start, goto_len);
     align_code();
     last_jump=0;
   }
 }
 
+static void append_jump_previous(void)
+/* append a dispatch to the previous primitive */
+{
+  append_jump();
+}
+
 /* Gforth remembers all code blocks in this list.  On forgetting (by
 executing a marker) the code blocks are not freed (because Gforth does
 not remember how they were allocated; hmm, remembering that might be
@@ -1188,6 +1246,9 @@ needed. */
 
 struct code_block_list {
   struct code_block_list *next;
+  /* long dynamicinfos; */
+     /* index into dynamicinfos for first primitive in block
+        intended for future faster implementation of dynamic_info() */
   Address block;
   Cell size;
 } *code_block_list=NULL, **next_code_blockp=&code_block_list;
@@ -1225,9 +1286,8 @@ static Address append_prim(Cell p)
   if(reserve_code_space(pi->length+pi->restlength+goto_len+CODE_ALIGNMENT-1))
     return NULL;
   /* debugp(stderr, "Copy code %p<=%p,%d\n", code_here, pi->start, pi->length); */
-  memmove(code_here, pi->start, pi->length);
   old_code_here = code_here;
-  code_here += pi->length;
+  append_code(pi->start, pi->length);
   return old_code_here;
 }
 
@@ -1262,6 +1322,9 @@ int forget_dyncode(Address code)
 #else
   struct code_block_list *p, **pp;
 
+  DynamicInfo *di = dynamic_info((Label)code);
+  if (di != NULL)
+    ndynamicinfos = di-dynamicinfos;
   for (pp=&code_block_list, p=*pp; p!=NULL; pp=&(p->next), p=*pp) {
     if (code >= p->block && code < p->block+p->size) {
       next_code_blockp = &(p->next);
@@ -1292,29 +1355,47 @@ static long dyncodesize(void)
 
 Label decompile_code(Label _code)
 {
-#ifdef NO_DYNAMIC
-  return _code;
-#else /* !defined(NO_DYNAMIC) */
-  Cell i;
-  struct code_block_list *p;
-  Address code=_code;
+  DynamicInfo *di = dynamic_info(_code);
+  if (di==NULL)
+    return _code;
+  return vm_prims[super2[super_costs[di->prim].offset]];
+}
 
-  /* first, check if we are in code at all */
-  for (p = code_block_list;; p = p->next) {
-    if (p == NULL)
-      return code;
-    if (code >= p->block && code < p->block+p->size)
-      break;
+static Cell prim_index(Label code)
+/* !! use binary search or hashing instead */
+{
+  long i;
+  for (i=0; vm_prims[i]!=NULL; i++)
+    if (vm_prims[i]==code)
+      return i;
+  return -1;
+}
+
+DynamicInfo *decompile_prim1(Label _code)
+{
+  DynamicInfo *di = dynamic_info(_code);
+  /* fprintf(stderr,"\n%p n=%ld\n",di,ndynamicinfos);*/
+  if (di==NULL) {
+    static DynamicInfo dyninfo; 
+    Cell p = prim_index(_code);
+    if (p<0)
+      dyninfo = (DynamicInfo){_code,-1,0,0,0,0};
+    else {
+      struct cost *c = &super_costs[p];
+      dyninfo = (DynamicInfo){_code,0,p,c->state_in,c->state_out};
+      assert(c->state_in  == CANONICAL_STATE);
+      assert(c->state_out == CANONICAL_STATE);
+    }
+    di = &dyninfo;
   }
-  /* reverse order because NOOP might match other prims */
-  for (i=npriminfos-1; i>DOER_MAX; i--) {
-    PrimInfo *pi=decomp_prims[i];
-    if (pi->start==code || (pi->start && memcmp(code,pi->start,pi->length)==0))
-      return vm_prims[super2[super_costs[pi-priminfos].offset]];
-    /* return pi->start;*/
-  }
-  return code;
-#endif /* !defined(NO_DYNAMIC) */
+  return di;
+}
+
+void update_dynamic_info()
+/* after appending something, include it in the last dynamicinfo */
+{
+  DynamicInfo *di = &dynamicinfos[ndynamicinfos-1];
+  di->length = code_here - (Address)di->start;
 }
 
 void finish_code(void)
@@ -1348,7 +1429,7 @@ static Cell compile_prim_dyn(PrimNum p, Cell *tcp)
     return static_prim;
   priminfos[p].uses++;
   if (p>=npriminfos || !is_relocatable(p)) {
-    append_jump();
+    append_jump_previous();
     return static_prim;
   }
   old_code_here = append_prim(p);
@@ -1419,7 +1500,6 @@ void init_ss_cost(void) {
 
 #define MAX_BB 128 /* maximum number of instructions in BB */
 #define INF_COST 1000000 /* infinite cost */
-#define CANONICAL_STATE 0
 
 struct waypoint {
   int cost;     /* the cost from here to the end */
@@ -1511,8 +1591,8 @@ static struct tpa_state *make_termstate()
 struct tpa_entry {
   struct tpa_entry *next;
   PrimNum inst;
-  struct tpa_state *state_behind;  /* note: brack-to-front labeling */
-  struct tpa_state *state_infront; /* note: brack-to-front labeling */
+  struct tpa_state *state_behind;  /* note: back-to-front labeling */
+  struct tpa_state *state_infront; /* note: back-to-front labeling */
 } *tpa_table[TPA_SIZE];
 
 #if !(defined(DOUBLY_INDIRECT) || defined(INDIRECT_THREADED))
@@ -1619,9 +1699,9 @@ static void optimize_rewrite(Cell *instps[], PrimNum origs[], int ninsts)
 {
   int i,j;
   struct tpa_state *ts[ninsts+1];
-  int nextdyn, nextstate, no_transition;
+  int nextdyn, nextstate, startstate, no_transition;
   Address old_code_area;
-  
+  DynamicInfo *di = NULL;
   lb_basic_blocks++;
   ts[ninsts] = termstate;
 #ifndef NO_DYNAMIC
@@ -1692,19 +1772,21 @@ static void optimize_rewrite(Cell *instps[], PrimNum origs[], int ninsts)
 		   ||ts[0]->trans[nextstate].no_transition);
   for (i=0; i<ninsts; i++) {
     Cell tc=0, tc2;
+    startstate = nextstate;
     if (i==nextdyn) {
+      PrimNum p;
       if (!no_transition) {
 	/* process trans */
-	PrimNum p = ts[i]->trans[nextstate].inst;
-	struct cost *c = super_costs+p;
+	PrimNum pt = ts[i]->trans[nextstate].inst;
+	struct cost *c = super_costs+pt;
 	assert(ts[i]->trans[nextstate].cost != INF_COST);
 	assert(c->state_in==nextstate);
-	tc = compile_prim_dyn(p,NULL);
+	tc = compile_prim_dyn(pt,NULL);
 	nextstate = c->state_out;
       }
+      p = ts[i]->inst[nextstate].inst;
       {
 	/* process inst */
-	PrimNum p = ts[i]->inst[nextstate].inst;
 	struct cost *c=super_costs+p;
 	assert(c->state_in==nextstate);
 	assert(ts[i]->inst[nextstate].cost != INF_COST);
@@ -1712,20 +1794,35 @@ static void optimize_rewrite(Cell *instps[], PrimNum origs[], int ninsts)
 	assert(p == origs[i]);
 #endif
 	tc2 = compile_prim_dyn(p,instps[i]);
-	if (no_transition || !is_relocatable(p))
+	if (no_transition || !is_relocatable(p)) {
 	  /* !! actually what we care about is if and where
 	   * compile_prim_dyn() puts NEXTs */
 	  tc=tc2;
+          startstate=nextstate;
+        }
 	no_transition = ts[i]->inst[nextstate].no_transition;
 	nextstate = c->state_out;
 	nextdyn += c->length;
       }
+      if (is_relocatable(p)) {
+        di = add_dynamic_info();
+        if (ndynamicinfos>1 &&
+            ((UCell)(((Address)tc)-code_area)) < (UCell)code_area_size) {
+          di[-1].length = ((Address)tc) - (Address)di[-1].start;
+          di[-1].end_state = startstate;
+        }
+        di->prim = p;
+        di->seqlen = super_costs[p].length;
+        di->start = (Label)tc;
+        di->length = code_here - (Address)di->start;
+        di->start_state = startstate;
+        di->end_state = nextstate;
+      }
     } else {
 #if defined(GFORTH_DEBUGGING)
       assert(0);
 #endif
-      tc=0;
-      /* tc= (Cell)vm_prims[ts[i]->inst[CANONICAL_STATE].inst]; */
+      tc= (Cell)vm_prims[super2[super_costs[ts[i]->inst[CANONICAL_STATE].inst].offset]];
     }
     *(instps[i]) = tc;
   }      
@@ -1737,6 +1834,8 @@ static void optimize_rewrite(Cell *instps[], PrimNum origs[], int ninsts)
     assert(i==nextdyn);
     (void)compile_prim_dyn(p,NULL);
     nextstate = c->state_out;
+    di->length = code_here - (Address)di->start;
+    di->end_state = nextstate;
   }
   assert(nextstate==CANONICAL_STATE);
   assert(code_area==old_code_area); /* does reserve_code_super() work? */
@@ -2296,7 +2395,7 @@ int gforth_args(int argc, char ** argv, char ** path, char ** imagename)
       {"ignore-async-signals", no_argument, &ignore_async_signals, 1},
       {"no-super", no_argument, &no_super, 1},
       {"no-dynamic", no_argument, &no_dynamic, 1},
-      {"no-rc0", no_argument, &no_rc0, 1},
+      {"no-0rc", no_argument, &no_rc0, 1},
       {"dynamic", no_argument, &no_dynamic, 0},
       {"code-block-size", required_argument, NULL, opt_code_block_size},
       {"print-metrics", no_argument, &print_metrics, 1},
@@ -2371,7 +2470,7 @@ Engine Options:\n\
   --no-dynamic			    Use only statically compiled primitives\n\
   --no-offset-im		    Load image at normal position\n\
   --no-super			    No dynamically formed superinstructions\n\
-  --no-rc0			    do not load ~/.config/gforthrc0\n\
+  --no-0rc			    do not load ~/.config/gforthrc0\n\
   --offset-image		    Load image at a different position\n\
   -p PATH, --path=PATH		    Search path for finding image and sources\n\
   --print-metrics		    Print some code generation metrics on exit\n\
